/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * h3600 atmel micro companion support 
 * based on previous kernel 2.4 version 
 * Author : Alessandro Gardich <gremlin@gremlin.it>
 *
 */


#include <linux/module.h>
#include <linux/version.h>

#include <linux/init.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/pm.h>
#include <linux/sysctl.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/device.h>
#include <linux/input.h>
#include <linux/platform_device.h>
#include <linux/backlight.h>    /* backlight_device */

#include <asm/irq.h>
#include <asm/arch/hardware.h>

#include <asm/arch/h3600.h>
#include <asm/arch/SA-1100.h>

#include "../../mach-sa1100/generic.h"

#define TX_BUF_SIZE	32
#define RX_BUF_SIZE	16
#define CHAR_SOF	0x02

/* state of receiver parser */
enum rx_state {
	STATE_SOF = 0,     /* Next byte should be start of frame */
	STATE_ID,          /* Next byte is ID & message length   */
	STATE_DATA,        /* Next byte is a data byte           */
	STATE_CHKSUM       /* Next byte should be checksum       */
};

unsigned char tx_buf[TX_BUF_SIZE];
unsigned char *tx_index=tx_buf;
int	tx_len=0;

struct h3600_rxdev {
        enum rx_state state;            /* context of rx state machine */
        unsigned char chksum;           /* calculated checksum */
        int           id;               /* message ID from packet */
        unsigned int  len;              /* rx buffer length */
        unsigned int  index;            /* rx buffer index */
        unsigned char buf[RX_BUF_SIZE]; /* rx buffer size  */
};

static struct h3600_rxdev rx;	/* receive ISR state */

/* forward declaration */
static void h3600_micro_tx_chars( void );


/*--- backlight ---*/
static struct backlight_device *micro_backlight_device;

#define MAX_BRIGHTNESS 255

int micro_backlight_level = 0;
int micro_backlight_brightness = 0;

static void micro_backlight_send (void)
{
	tx_buf[0] = 0x02;
	tx_buf[1] = 0xD3;
	tx_buf[2] = 0x01;
	tx_buf[3] = micro_backlight_level;
	tx_buf[4] = micro_backlight_brightness;
	tx_buf[5] = (0xD4 + micro_backlight_level + micro_backlight_brightness) & 0xff;
	tx_len=6;
	
	h3600_micro_tx_chars();
}

static void __micro_backlight_set_power(int level) 
{
	if (level > 0) {
		micro_backlight_level = 1;
	} else {
		micro_backlight_level = 0;
	}

	micro_backlight_send();
}


static int micro_backlight_set_power (struct backlight_device *bl, int level)
{
	__micro_backlight_set_power(level);
	return 0;
}

static int micro_backlight_get_power (struct backlight_device *bl)
{
	return micro_backlight_level;
}

int micro_backlight_set_brightness (struct backlight_device *bl, int brightness)
{
	micro_backlight_brightness = brightness;
	if (brightness > MAX_BRIGHTNESS)
		micro_backlight_brightness = MAX_BRIGHTNESS;
	if (brightness < 0) 
		micro_backlight_brightness = 0;

	micro_backlight_send();

	return 0;
}

int micro_backlight_get_brightness (struct backlight_device *bl)
{
	return micro_backlight_brightness;
}

static struct backlight_properties micro_backlight_properties = {
	.owner		= THIS_MODULE,
	.get_power	= micro_backlight_get_power,
	.set_power	= micro_backlight_set_power,
	.get_brightness = micro_backlight_get_brightness,
	.set_brightness = micro_backlight_set_brightness,
	.max_brightness = 0xff,
};

/*--- manage messages from Atmel Micro  ---*/
typedef struct { 
	char	test[8];
	void    (*handle)( int len, unsigned char *data );
} micro_private_t;

static micro_private_t micro;

static void h3600_micro_rx_msg( int id, int len, unsigned char *data )
{
	int i;

	/*printk(KERN_ERR "h3600_micro : got a message from micro\n");*/
	switch (id) { 
	case 0x0D : /* backlight */
		/* empty ack, just ignore */
		break;
	case 0x02 : /* keyboard */
		/*printk(KERN_ERR "h3600_micro : key message\n");*/
		if (micro.handle != NULL) {
			micro.handle(len, data);
			printk(KERN_ERR "h3600_micro : key message %s ",micro.test);
		} else {
			printk(KERN_ERR "h3600_micro : key message ignored, no handle ");
		}
		break;
	default : 
		printk(KERN_ERR "h3600_micro : unknown msg %d [%d] ", id, len);
		for (i=0;i<len;++i)
			printk("0x%02x ", data[i]);
		printk("\n");
	}
}

/*--- low lever serial interface ---*/

static void h3600_micro_process_char( unsigned char ch )
{
	switch ( rx.state ) {
	case STATE_SOF:	/* Looking for SOF */
		if ( ch == CHAR_SOF )
			rx.state=STATE_ID; /* Next byte is the id and len */
		//else 
		//	g_statistics.missing_sof++;
		break;
	case STATE_ID: /* Looking for id and len byte */
		rx.id = ( ch & 0xf0 ) >> 4 ;
		rx.len = ( ch & 0x0f );
		rx.index = 0;
		rx.chksum = ch;
		rx.state = ( rx.len > 0 ) ? STATE_DATA : STATE_CHKSUM;
		break;
	case STATE_DATA: /* Looking for 'len' data bytes */
		rx.chksum += ch;
		rx.buf[rx.index]= ch;
		if ( ++rx.index == rx.len )
			rx.state = STATE_CHKSUM;
		break;
	case STATE_CHKSUM: /* Looking for the checksum */
		if ( ch == rx.chksum )
			h3600_micro_rx_msg( rx.id, rx.len, rx.buf );
		//else
		//	g_statistics.bad_checksum++;
		rx.state = STATE_SOF;
		break;
	}
}

static void h3600_micro_rx_chars( void )
{
	unsigned int status, ch;

	while ( (status = Ser1UTSR1) & UTSR1_RNE ) {
		ch = Ser1UTDR;
		/*g_statistics.rx++;*/

		if ( status & UTSR1_PRE ) { /* Parity error */
			printk(KERN_ERR "h3600_micro_rx : parity error\n");
			/*g_statistics.parity++;*/
		} else 	if ( status & UTSR1_FRE ) { /* Framing error */
			printk(KERN_ERR "h3600_micro_rx : framing error\n");
			/*g_statistics.frame++;*/
		} else {
			if ( status & UTSR1_ROR ) {  /* Overrun error */
				printk(KERN_ERR "h3600_micro_rx : overrun error\n");
				/*g_statistics.overrun++;*/
			} 
			h3600_micro_process_char(ch);
		}
	}
}

static void h3600_micro_tx_chars( void )
{
	tx_index = tx_buf;	

	while ( (Ser1UTSR1 & UTSR1_TNF) ) {
		Ser1UTDR = *tx_index;
		tx_index++;
		tx_len--;
		if (tx_len==0) break;
	}

	/*if ( g_txdev.tail == g_txdev.head )*/ /* Stop interrupts */
	/*	Ser1UTCR3 &= ~UTCR3_TIE;     */
}

static void h3600_micro_reset_comm( void )
{
	if (1) printk("%s: initializing serial port\n", __FUNCTION__);

	/* Initialize Serial channel protocol frame */
	rx.state = STATE_SOF;  /* Reset the state machine */

	/* Set up interrupts */
	Ser1SDCR0 = 0x1;                               /* Select UART mode */

	Ser1UTCR3 = 0;                                 /* Clean up CR3                  */
	Ser1UTCR0 = UTCR0_8BitData | UTCR0_1StpBit;    /* 8 bits, no parity, 1 stop bit */
	Ser1UTCR1 = 0;                                 /* Baud rate to 115K bits/sec    */
	Ser1UTCR2 = 0x1;

	Ser1UTSR0 = 0xff;                              /* Clear SR0 */
	Ser1UTCR3 = UTCR3_TXE | UTCR3_RXE | UTCR3_RIE; /* Enable receive interrupt */
	Ser1UTCR3 &= ~UTCR3_TIE;                       /* disable send interrupt */
}

/*--- core interrupt ---*/
enum MessageHandleType {
	HANDLE_NORMAL,
	HANDLE_ACK,
	HANDLE_ERROR
};

#define MICRO_MSG_WAITING 0
#define MICRO_MSG_SUCCESS 1
#define MICRO_MSG_ERROR  -1

static irqreturn_t h3600_micro_serial_isr(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned int status; /* UTSR0 */
	/*unsigned int pass_counter = 0;*/

	if (1) printk("%s: %s\n", __FILE__, __FUNCTION__);

	//g_statistics.isr++;
	status = Ser1UTSR0;
	do {
		if ( status & (UTSR0_RID | UTSR0_RFS) ) {
			if ( status & UTSR0_RID )
				Ser1UTSR0 = UTSR0_RID; /* Clear the Receiver IDLE bit */
			h3600_micro_rx_chars();
		}

		/* Clear break bits */
		if (status & (UTSR0_RBB | UTSR0_REB))
			Ser1UTSR0 = status & (UTSR0_RBB | UTSR0_REB);

		/* at now i send immediatelly, to be modified */
		/*if ( status & UTSR0_TFS )     */
		/*	h3600_micro_tx_chars(); */

		status = Ser1UTSR0;
	} while ( ( /*( (g_txdev.head != g_txdev.tail) && (status & UTSR0_TFS) ) ||*/
	          status & (UTSR0_RFS | UTSR0_RID ) )
	          /*&& pass_counter++ < H3600_TS_PASS_LIMIT*/ );

	//if ( pass_counter >= H3600_TS_PASS_LIMIT )
	//	g_statistics.pass_limit++;

	return IRQ_HANDLED;
}

/*--- sub devices declaration ---*/
enum { h3600bl=0, h3600keys, n_devices };
static struct platform_device *devices[n_devices]; 

/*--- micro ---*/
static int micro_suspend (struct device *dev, u32 state, u32 level)
{
	micro_backlight_set_power (micro_backlight_device, 0);
	return 0;
}

static int micro_resume (struct device *dev, u32 level)
{
	micro_backlight_set_power (micro_backlight_device, 4);
	return 0;
}

static int micro_probe (struct device *dev)
{
	int result=0;
	struct platform_device *plat;
	
	printk(KERN_ERR "micro probe : begin \n");

	h3600_micro_reset_comm();

	result = request_irq(IRQ_Ser1UART, h3600_micro_serial_isr, 
	                     SA_SHIRQ | SA_INTERRUPT | SA_SAMPLE_RANDOM,
	                     "h3600_micro", h3600_micro_serial_isr);
	if ( result ) {
		printk(KERN_CRIT "%s: unable to grab serial port IRQ\n", __FUNCTION__);
		return result;
	} else { 
		printk(KERN_ERR "h3600_micro : grab serial port IRQ\n");
	}
	
	plat = to_platform_device (dev);
	plat->dev.platform_data = &micro; 
	micro.handle = NULL;

	printk(KERN_ERR "micro probe : end [%d]\n",result);

	return result;
}

static int micro_remove (struct device *dev)
{
	micro_backlight_set_power (micro_backlight_device, 0);
	mdelay(10);
	sa1100fb_backlight_power=NULL;
	backlight_device_unregister(micro_backlight_device);
	platform_device_unregister(devices[h3600bl]);
	platform_device_unregister(devices[h3600keys]);
	Ser1UTCR3 &= ~(UTCR3_RXE | UTCR3_RIE); /* disable receive interrupt */
	Ser1UTCR3 &= ~(UTCR3_TXE | UTCR3_TIE); /* disable send interrupt */
	free_irq(IRQ_Ser1UART, h3600_micro_serial_isr);
	return 0;
}

static struct device_driver micro_device_driver = {
	.name     = "h3600-micro",
	.bus      = &platform_bus_type,
	.probe    = micro_probe,
	.remove   = micro_remove,
	.suspend  = micro_suspend,
	.resume   = micro_resume,
	.shutdown = micro_suspend
};

static int micro_init (void)
{
	int result;
	result = driver_register (&micro_device_driver);
	if ( result ) {
		return result;
	}
	
	/*--- add platform devices that should be avaiable with micro ---*/
	devices[h3600bl] = platform_device_alloc("h3600-micro-bl", -1); 
	devices[h3600keys] = platform_device_alloc("h3600-micro-keys", -1); 
	platform_device_add_data(devices[h3600keys], &micro, sizeof(micro));
	
	result = platform_add_devices(devices, n_devices);
	if ( !result ) {
		printk(KERN_ERR "micro probe : platform_add_devices fail.\n");
	}

	/*--- register backlight ---*/
	micro_backlight_device = backlight_device_register ("h3600-micro-bl", NULL, 
	                                                    &micro_backlight_properties);
	if (IS_ERR (micro_backlight_device)) {
		printk(KERN_ERR "micro probe : backlight register fail.\n");
		return PTR_ERR (micro_backlight_device);
	}

	micro_backlight_set_power (micro_backlight_device, 1);
	sa1100fb_backlight_power=__micro_backlight_set_power;

	return result;
}

static void
micro_cleanup (void)
{
	driver_unregister (&micro_device_driver);
}

module_init (micro_init);
module_exit (micro_cleanup);

int h3600_micro_dummy_dep(void) { return 0; }
EXPORT_SYMBOL(h3600_micro_dummy_dep);


MODULE_LICENSE("GPL");
MODULE_AUTHOR("gremlin.it");
MODULE_DESCRIPTION("driver for iPAQ Atmel micro core and backlight");

