/*
 * Experimental driver for the CPLD in the HTC Magician
 *
 * Copyright 2006 Philipp Zabel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/module.h>
#include <linux/version.h>

#include <linux/interrupt.h>
/*#include <linux/irq.h>*/
#include <linux/platform_device.h>

#include <asm/io.h>
#include <asm/irq.h>

#include <asm/arch/hardware.h>
#include <asm/arch/pxa-regs.h>

#include <asm/arch/magician.h>

struct magician_cpld_registers
{
	u32 egpio[3];
	u32 irq_state;
	u32 cable_state;
};

struct magician_cpld_data
{
	struct magician_cpld_registers *mapping;
	int irq_nr;
	u32 cached_egpio[3];	// cached egpio registers
};

static inline int u8pos(int bit)
{
	return bit/8;
}

static inline int u8bit(int bit)
{
	return 1<<(bit & (8-1));
}

void magician_egpio_enable (struct device *dev, int bit)
{
	unsigned long flags;
	struct magician_cpld_data *cpld = (struct magician_cpld_data *)dev->driver_data;
	int pos = u8pos(bit);

	local_irq_save(flags);

	cpld->cached_egpio[pos] |= u8bit(bit);
	printk("egpio set: reg %d = 0x%08x\n", pos, cpld->cached_egpio[pos]);
	cpld->mapping->egpio[pos] = cpld->cached_egpio[pos];

	local_irq_restore(flags);
}
EXPORT_SYMBOL(magician_egpio_enable);

void magician_egpio_disable (struct device *dev, int bit)
{
	unsigned long flags;
	struct magician_cpld_data *cpld = (struct magician_cpld_data *)dev->driver_data;
	int pos = u8pos(bit);

	local_irq_save(flags);

	cpld->cached_egpio[pos] &= ~u8bit(bit);
	printk("egpio set: reg %d = 0x%08x\n", pos, cpld->cached_egpio[pos]);
	cpld->mapping->egpio[pos] = cpld->cached_egpio[pos];

	local_irq_restore(flags);
}
EXPORT_SYMBOL(magician_egpio_disable);

#if 0
static void magician_cpld_ack_gpio(unsigned int irq)
{
//	GEDR0 = (1 << (irq - IRQ_GPIO0));
	printk ("magician_cpld_ack_gpio(%d)\n", irq);
}

static void magician_cpld_mask_irq(unsigned int irq)
{
}

static void magician_cpld_unmask_irq(unsigned int irq)
{
}

static struct irqchip magician_cpld_chip = {
	.ack		= magician_cpld_ack_gpio,
	.mask		= magician_cpld_mask_irq,
	.unmask		= magician_cpld_unmask_irq,
};

/*
 * Demux handler for CPLD edge detect interrupts
 */

static void magician_cpld_demux_handler(unsigned int irq,
			struct irqdesc *desc, struct pt_regs *regs)
{
	struct magician_cpld_data *cpld = desc->chip_data;
	int result, cable_state;

//	irq = USBC
//	desc = irq_desc + irq;
//	desc_handle_irq(irq, desc, regs);

	result = cpld->mapping->irq_state;
	printk ("0x0c00000c = 0x%08x\n", result);
	result = cpld->mapping->irq_state;
	printk ("0x0c00000c = 0x%08x\n", result);
	cpld->mapping->irq_state = result;			// ACK

	cable_state = cpld->mapping->cable_state;
	printk ("cable_state = 0x%08x\n", cable_state);		// pCPLDRegs->cable_state

	switch (result) {
		case 0x1:
			irq = IRQ_MAGICIAN_SD;
			desc = &irq_desc[irq];
			desc_handle_irq(irq, desc, regs);
			break;
		case 0x2:
			irq = IRQ_MAGICIAN_EP;
			desc = &irq_desc[irq];
			desc_handle_irq(irq, desc, regs);
			break;
		case 0x8:
			irq = IRQ_MAGICIAN_AC;
			desc = &irq_desc[irq];
			desc_handle_irq(irq, desc, regs);
			break;
		case 0xc:
			irq = IRQ_MAGICIAN_USBC;
			desc = &irq_desc[irq];
			desc_handle_irq(irq, desc, regs);
			break;
		default:
			printk ("magician_cpld: unknown irq 0x%x\n", result);
	}
}
#endif

static int magician_cpld_probe (struct device *dev)
{
	struct platform_device *pdev = to_platform_device (dev);
//	struct magician_cpld_platform_data *pdata = dev->platform_data;
	struct magician_cpld_data *cpld;

	cpld = kmalloc (sizeof (struct magician_cpld_data), GFP_KERNEL);
	if (!cpld)
		return -ENOMEM;
	memset (cpld, 0, sizeof (*cpld));

	dev->driver_data = cpld;

	cpld->mapping = (struct magician_cpld_registers *)ioremap ((unsigned long)pdev->resource[0].start, 0x1000);
	if (!cpld->mapping) {
		printk ("magician_cpld: couldn't ioremap\n");
		kfree (cpld);
		return -ENOMEM;
	}

	/*
	 * Initial values. FIXME: if we're not booting from wince, we should set them ourselves.
	 */
	cpld->cached_egpio[0] = cpld->mapping->egpio[0];
	cpld->cached_egpio[1] = cpld->mapping->egpio[1];
	cpld->cached_egpio[2] = cpld->mapping->egpio[2];
	printk ("initial egpios: 0x%08x 0x%08x 0x%08x\n", cpld->cached_egpio[0], cpld->cached_egpio[1], cpld->cached_egpio[2]);

#if 0
	if (pdev->num_resources > 1)
		cpld->irq_nr = pdev->resource[1].start;

	if (cpld->irq_nr != -1) {
		unsigned int irq;

		/*
		 * Setup handlers for CPLD irqs, muxed into IRQ_GPIO(13)
		 */
		for (irq = IRQ_MAGICIAN_SD; irq <= IRQ_MAGICIAN_AC; irq++) {
			set_irq_chip(irq, &magician_cpld_chip);
			set_irq_chipdata(irq, cpld);
			set_irq_handler(irq, do_simple_IRQ);
			set_irq_flags(irq, IRQF_VALID | IRQF_PROBE); // | ACK
		}

		set_irq_chained_handler(irq, magician_cpld_demux_handler);
		set_irq_type (irq, IRQT_RISING);
		set_irq_data (cpld->irq_nr, cpld);

		if (!(GRER0 & 0x2000)) {
			printk ("FIXME: setting GRER0|=0x2000, shouldn't the set_irq_type above have done this?");
			GRER0 |= 0x2000;
			GFER0 &= ~0x2000;
		}
	}
#endif
	return 0;
}

static int magician_cpld_remove (struct device *dev)
{
	struct magician_cpld_data *cpld;
//	struct magician_cpld_platform_data *pdata = dev->platform_data;
	cpld = dev->driver_data;
#if 0
	if (cpld->irq_nr != -1) {
		unsigned int irq;

		for (irq = IRQ_MAGICIAN_SD; irq <= IRQ_MAGICIAN_AC; irq++) {
			set_irq_chip(irq, NULL);
			set_irq_chipdata(irq, NULL);
			set_irq_handler(irq, NULL);
			set_irq_flags(irq, 0);
		}

		set_irq_chained_handler(cpld->irq_nr, NULL);

		if ((GRER0 & 0x2000)) {
			printk ("FIXME: shouldn't this be done with set_irq_type?");
			GRER0 &= ~0x2000;
			GFER0 &= ~0x2000;
		}
	}
#endif
	iounmap((void *)cpld->mapping);

	return 0;
}


struct device_driver magician_cpld_device_driver = {
	.name     = "magician-cpld",
	.bus	  = &platform_bus_type,
	.probe    = magician_cpld_probe,
	.remove   = magician_cpld_remove,
//	.suspend  =
//	.resume   =
};

static int magician_cpld_init (void)
{
	return driver_register (&magician_cpld_device_driver);
}

static void magician_cpld_exit (void)
{
	driver_unregister (&magician_cpld_device_driver);
}

module_init (magician_cpld_init);
module_exit (magician_cpld_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Philipp Zabel <philipp.zabel@gmail.com>");
MODULE_DESCRIPTION("Experimental driver for the CPLD in the HTC Magician");
MODULE_SUPPORTED_DEVICE("magician-cpld");
