/*
 * LCD driver for HTC Magician
 *
 * Copyright (C) 2006, Philipp Zabel
 *
 * based on previous work by
 *	Giuseppe Zompatori <giuseppe_zompatori@yahoo.it>,
 *	Andrew Zabolotny <zap@homelink.ru>
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of this archive for
 * more details.
 */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/notifier.h>
#include <linux/lcd.h>
#include <linux/err.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/platform_device.h>

#include <asm/arch/magician.h>
#include <asm/mach-types.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/pxafb.h>

#include <asm/arch/magician_cpld.h>

static struct pxafb_mach_info magician_fb_info1 = {
	.pixclock = 96153,	// TODO: check that LCCR4=0x80000000 ( 104Mhz/(k=2)/10kHz*pixclock/2/100*1e6-1=pcd=0x5.)
	.bpp = 16,		//
	.xres = 240,		// PPL + 1 0x1df
	.yres = 320,		// LPP + 1 0x27f
	.hsync_len = 4,		// HSW + 1 0x3
	.vsync_len = 1,		// VSW + 1 0x0
	.left_margin = 20,	// BLW + 1 0x13
	.upper_margin = 7,	// BFW     0x07
	.right_margin = 8,	// ELW + 1 0x07
	.lower_margin = 8,	// EFW 0x08
	.sync = 0,		// ???
	.lccr0 = 0x00000080,	// yes that's ugly, but it's due a limitation of the
//      .lccr3 = 0x04700005, // pxafb driver assuming pxa25x fbs...
	.lccr3 = 0x00400000,	// bpp and pcd removed. TODO: use the values defined in pxafb.h
	.pxafb_backlight_power = NULL,
	.pxafb_lcd_power = NULL,
};

static struct pxafb_mach_info magician_fb_info2 = {
	.pixclock = 96153,	// TODO: check that LCCR4=0x80000000 ( 104Mhz/(k=2)/10kHz*pixclock/2/100*1e6-1=pcd=0x5.)
//PCD=0x4
	.bpp = 16,		//
	.xres = 240,		// PPL + 1 0x1df
	.yres = 320,		// LPP + 1 0x27f
	.hsync_len = 11,	// HSW + 1 0x3
	.vsync_len = 3,		// VSW + 1 0x0
	.left_margin = 19,	// BLW + 1 0x13
	.upper_margin = 2,	// BFW     0x07
	.right_margin = 10,	// ELW + 1 0x07
	.lower_margin = 2,	// EFW 0x08
	.sync = 0,		// ???
// wince: 00000081 (LCCR0_Pas = 0x80, LCCR0_ENB = 0x1)
	.lccr0 = 0x00000080,	// enb removed, see pxafb.h
// wince: 0x04300004 (BPP=4, HSP,VSP, PCD=4)
	.lccr3 = 0x00000000,	// bpp, hsp, vsp and pcd removed
	.pxafb_backlight_power = NULL,
	.pxafb_lcd_power = NULL,
};

/*
 * pH5's magician needs info2. awelux' magician needs info1
 * one is supposed to have a Samsung LTP280QV, the other a
 * Toppoly TD028STEB1 lcd.
 * The bootloader checks ldrb r0, [0xa9b00014] (r0 & 0x8)
 */
#define magician_fb_info magician_fb_info2

extern struct platform_device magician_cpld;

static int lcd_power;

static int magician_lcd_set_power(struct lcd_device *lm, int level)
{
	printk("magician_lcd_set_power(level=%d) called.\n", level);

	switch (level) {
	case FB_BLANK_UNBLANK:
//		magician_egpio_enable(&magician_cpld.dev, EGPIO_NR_MAGICIAN_LCD_POWER);
		break;
	case FB_BLANK_VSYNC_SUSPEND:
	case FB_BLANK_HSYNC_SUSPEND:
		break;
	case FB_BLANK_NORMAL:
	case FB_BLANK_POWERDOWN:
//		magician_egpio_disable(&magician_cpld.dev, EGPIO_NR_MAGICIAN_LCD_POWER);
		break;
	}

	lcd_power = level;

	return 0;
}

static int magician_lcd_get_power(struct lcd_device *lm)
{
	return lcd_power;
}

struct lcd_properties magician_lcd_properties = {
	.owner = THIS_MODULE,
	.set_power = magician_lcd_set_power,
	.get_power = magician_lcd_get_power,
};

static struct lcd_device *pxafb_lcd_device;

struct pm_save_data {
	int brightness;
};

static int magician_lcd_probe(struct device *dev)
{
	int lcd_select;
	volatile u8 *p;

	printk(KERN_NOTICE "HTC Magician LCD driver\n");

	set_pxa_fb_info(&magician_fb_info);

	pxafb_lcd_device =
	    lcd_device_register("pxafb", NULL, &magician_lcd_properties);
	if (IS_ERR(pxafb_lcd_device))
		return PTR_ERR(pxafb_lcd_device);

	/* Check which LCD we have */
	p = (volatile u8 *)ioremap_nocache(PXA_CS3_PHYS, 0x1000);
	if (p) {
		lcd_select = p[14];
		printk("magician_lcd: lcd_select = %d (%s)\n", lcd_select,
		       (lcd_select & 0x8) ? "samsung" : "toppoly");
		iounmap(p);
	} else
		printk("magician_lcd: lcd_select ioremap failed\n");

	return 0;
}

static int magician_lcd_remove(struct device *dev)
{
	lcd_device_unregister(pxafb_lcd_device);

	return 0;
}

#ifdef CONFIG_PM
static int magician_lcd_suspend(struct device *dev, pm_message_t state)
{
	struct pm_save_data *save;

	save = kmalloc(sizeof(struct pm_save_data), GFP_KERNEL);
	if (!save)
		return -ENOMEM;

	dev->power.saved_state = save;

	return 0;
}

static int magician_lcd_resume(struct device *dev)
{
	struct pm_save_data *save;

	save = (struct pm_save_data *)dev->power.saved_state;
	if (!save)
		return 0;

	return 0;
}
#endif

static struct device_driver magician_lcd_driver = {
	.name = "magician-lcd",
	.bus = &platform_bus_type,
	.probe = magician_lcd_probe,
	.remove = magician_lcd_remove,
#ifdef CONFIG_PM
	.suspend = magician_lcd_suspend,
	.resume = magician_lcd_resume,
#endif
};

static __init int magician_lcd_init(void)
{
	if (!machine_is_magician())
		return -ENODEV;

	return driver_register(&magician_lcd_driver);
}

static __exit void magician_lcd_exit(void)
{
	driver_unregister(&magician_lcd_driver);
}

module_init(magician_lcd_init);
module_exit(magician_lcd_exit);

MODULE_AUTHOR("Philipp Zabel");
MODULE_DESCRIPTION("HTC Magician LCD driver");
MODULE_LICENSE("GPL");
