/*
 * Driver interface for keys on PXA25x GPIO lines
 *
 * Copyright 2005 Phil Blundell
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/version.h>

#include <linux/init.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/pm.h>
#include <linux/sysctl.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/device.h>
#include <linux/platform_device.h>
#include <linux/input.h>

#include <asm/irq.h>
#include <asm/arch/hardware.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/pxa_keys.h>

static irqreturn_t
pxa_keys_isr (int irq, void *dev_id, struct pt_regs *regs)
{
	int i;
	struct pxa_keys_platform_data *k = dev_id;

	for (i = 0; i < k->nbuttons; i++) {
		int gpio = k->buttons[i].gpio;
		if (irq == IRQ_GPIO(gpio)) {
			int state;
			state = ((GPLR(gpio) & GPIO_bit(gpio)) ? 1 : 0) ^ (k->buttons[i].active_low);
			if (state != k->buttons[i].down) {
				input_report_key (k->input, k->buttons[i].keycode, state);
				k->buttons[i].down = state;
			}
		}
	}

	return IRQ_HANDLED;
}

static int
pxa_keys_probe (struct device *dev)
{
	struct pxa_keys_platform_data *k;
	struct platform_device *pdev;
	int i;

	pdev = to_platform_device (dev);

	k = dev->platform_data;

	k->input = input_allocate_device();
	if (!k->input)
		return -ENOMEM;

	k->input->evbit[0] = BIT (EV_KEY);
	for (i = 0; i < k->nbuttons; i++) {
		int code = k->buttons[i].keycode;
		int irq = IRQ_GPIO (k->buttons[i].gpio);
		int result;

		set_irq_type (irq, IRQT_BOTHEDGE);
		result = request_irq (irq, pxa_keys_isr, SA_SAMPLE_RANDOM,
				      "pxa_keys", k);
		if (result == 0)
			set_bit (code, k->input->keybit);
		else
			printk("pxa_keys: unable to claim irq %d; error %d\n", irq, result);
	}

	k->input->name = pdev->name;
	k->input->private = k;

	input_register_device (k->input);

	return 0;
}

static int
pxa_keys_remove (struct device *dev)
{
	struct pxa_keys_platform_data *k = dev->platform_data;
	int i;
	
	for (i = 0; i < k->nbuttons; i++) {
		int irq = IRQ_GPIO (k->buttons[i].gpio);
		free_irq (irq, k);
	}

	input_unregister_device (k->input);

	return 0;
}

struct device_driver pxa_keys_device_driver = {
	.name     = "pxa2xx-keys",
	.bus	  = &platform_bus_type,
	.probe    = pxa_keys_probe,
	.remove   = pxa_keys_remove,
};

static int
pxa_keys_init (void)
{
	return driver_register (&pxa_keys_device_driver);
}

static void
pxa_keys_cleanup (void)
{
	driver_unregister (&pxa_keys_device_driver);
}

module_init (pxa_keys_init);
module_exit (pxa_keys_cleanup);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Phil Blundell <pb@handhelds.org>");
MODULE_DESCRIPTION("Keyboard driver for PXA25x/PXA27x GPIOs");
