/*
 * linux/arch/arm/common/t7l66xb_soc.c
 *
 * Toshiba T7L66XB support
 * Copyright (c) 2005 Ian Molton
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This file contains all generic T7L66XB support.
 *
 * All initialization functions provided here are intended to be called
 * from machine specific code with proper arguments when required.
 *
 * This is based on Dirk and my work on the tc6393xb SoC.
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/device.h>
#include <linux/slab.h>
#include <linux/spinlock.h>
#include <linux/soc-device.h>
#include <linux/soc/tmio_mmc.h>

#include <asm/hardware.h>
#include <asm/mach-types.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/mach/irq.h>
#include <asm/arch/pxa-regs.h>

#include <asm/hardware/t7l66xb.h>

struct t7l66xb_data
{
	int irq_base, irq_nr;
	void *mapbase;
	struct platform_device *devices;
	int ndevices;
};

struct t7l66xb_block
{
	int id;
	char *name;
	unsigned long ctl_start, ctl_end;
	unsigned long cnf_start, cnf_end;
	int irq;
	void *hwconfig;
};

static struct t7l66xb_block t7l66xb_blocks[] = {
	{
		.name  = "tmio_mmc",
		.id    = TMIO_MMC_DEVICE_ID,
		.cnf_start = T7L66XB_MMC_CONF_BASE,
		.cnf_end   = T7L66XB_MMC_CONF_BASE + 0xff,
		.ctl_start = T7L66XB_MMC_CTL_BASE,
		.ctl_end   = T7L66XB_MMC_CTL_BASE + 0x1ff,
		.irq       = T7L66XB_MMC_IRQ,
//		.hwconfig  = &t7l66xb_mmc_hwconfig,
	},
/*
	{
		.id    = TMIO_USB_DEVICE_ID,
		.name  = "tmio_ohci",
		.ctl_start = TC6393_USB_CTL_BASE,
		.ctl_end   = TC6393_USB_CTL_BASE + 0xff,
		.irq   = -1,
	},
*/
};


/* Handle the T7L66XB interrupt mux */
static void t7l66xb_irq_handler(unsigned int irq, struct irqdesc *desc,
			struct pt_regs *regs)
{
	int req, i;
	struct irqdesc *d = NULL;
	struct t7l66xb_data *data;
	
	data = desc->data;
	
	/* Acknowledge the parent IRQ */
	desc->chip->ack(irq);
  
	while ( (req = (readb(data->mapbase + T7L66XB_SYS_ISR)
			& ~(readb(data->mapbase + T7L66XB_SYS_IMR)))) ) {
		for (i = 0; i <= 7; i++) {
			if ((req & (1<<i))) {
				d = irq_desc + (data->irq_base + i);
				d->handle(data->irq_base + i, d, regs);
			}
		}
	}
}


static void t7l66xb_mask_irq(unsigned int irq)
{
	struct t7l66xb_data *data = get_irq_chipdata(irq);

	writeb(readb(data->mapbase + T7L66XB_SYS_IMR) | 1 << (irq - data->irq_base), data->mapbase + T7L66XB_SYS_IMR);
}

static void t7l66xb_unmask_irq(unsigned int irq)
{
	struct t7l66xb_data *data = get_irq_chipdata(irq);

	writeb(readb(data->mapbase + T7L66XB_SYS_IMR) & ~( 1 << (irq - data->irq_base)),data->mapbase + T7L66XB_SYS_IMR);
}

static struct irqchip t7l66xb_chip = {
	.ack	= t7l66xb_mask_irq,
	.mask	= t7l66xb_mask_irq,
	.unmask	= t7l66xb_unmask_irq,
};

/* Install the IRQ handler */
static void t7l66xb_setup_irq(struct t7l66xb_data *tchip)
{
	int i;
	
	for (i = 0; i < T7L66XB_NR_IRQS; i++) {
		int irq = tchip->irq_base + i;
		irq_desc[irq].valid    = 1;
		irq_desc[irq].probe_ok = 1;
		set_irq_chip (irq, &t7l66xb_chip);
		set_irq_chipdata (irq, tchip);
		set_irq_handler (irq, do_level_IRQ);
		set_irq_flags(irq, IRQF_VALID | IRQF_PROBE);
	}

	set_irq_chained_handler (tchip->irq_nr, t7l66xb_irq_handler);
	set_irq_type (tchip->irq_nr, IRQT_FALLING);
	set_irq_data (tchip->irq_nr, tchip);
}

static void t7l66xb_hwinit(struct device *dev)
{
	struct t7l66xb_platform_data *pdata = (struct t7l66xb_platform_data *)dev->platform_data;

	if (pdata->hw_init)
		pdata->hw_init();
		
}


#ifdef CONFIG_PM

static int t7l66xb_suspend(struct device *dev, u32 state, u32 level)
{
	struct t7l66xb_platform_data *pdata = (struct t7l66xb_platform_data *)dev->platform_data;

	if (level != SUSPEND_DISABLE)
		return 0;

	if (pdata->suspend)
		pdata->suspend();

	return 0;
}

static int t7l66xb_resume(struct device *dev, u32 level)
{
	struct t7l66xb_platform_data *pdata = (struct t7l66xb_platform_data *)dev->platform_data;

	if (pdata->resume)
		pdata->resume();

	t7l66xb_hwinit(dev);

	return 0;
}
#endif

static void t7l66xb_remove_devices(struct t7l66xb_data *data){
	int i;

	if(!data->devices)
		return;

	for(i = 0 ; i <  data->ndevices; i++) {
                if(data->devices[i].resource) {
                        platform_device_unregister(&data->devices[i]);
                        kfree(data->devices[i].resource);
                }
        }
        kfree(data->devices);
}

static int t7l66xb_add_devices(struct t7l66xb_data *data, struct device *dev, unsigned long pbase) {
	int n_devs = ARRAY_SIZE(t7l66xb_blocks);
	int i;

        data->devices = kmalloc(n_devs * sizeof(struct platform_device), GFP_KERNEL);
        if(!data->devices)
		return 0;

	data->ndevices = n_devs;
        memset (data->devices, 0, n_devs * sizeof (struct platform_device));
	for (i = 0; i < n_devs; i++) {
                struct platform_device *sdev = &data->devices[i];
                struct t7l66xb_block *blk = &t7l66xb_blocks[i];
                struct resource *res;

                /* Allocat the subdevices platform_device struct */
                sdev->device_id = blk->id;
                sdev->name = blk->name;
		printk("T7L66XB: register %s\n", blk->name);
                sdev->device.parent = dev;
                sdev->device.platform_data = (void *)blk->hwconfig;
                sdev->num_resources = 1; // All devices have a CTL reg.
                sdev->num_resources += blk->cnf_start?1:0;
                sdev->num_resources += (blk->irq >= 0)?1:0;
//              sdev->device.release = release;

                /* Allocate space for the subdevice resources */
                res = kmalloc (sdev->num_resources * sizeof (struct resource), GFP_KERNEL);
                sdev->resource = res;
		if(!res) {
			t7l66xb_remove_devices(data);
			return 0;
		}
                memset (res, 0, sdev->num_resources * sizeof (struct resource));
                /* Setup the subdevice resources */
                res->start = pbase + blk->ctl_start;
                res->end   = pbase + blk->ctl_end;
                res->flags = IORESOURCE_MEM | IORESOURCE_PLATFORM_VIRTUAL;
		if(blk->cnf_start) {
                        res++;
                        res->start = pbase + blk->cnf_start;
                        res->end   = pbase + blk->cnf_end;
                        res->flags = IORESOURCE_MEM | IORESOURCE_PLATFORM_VIRTUAL;
                }
                if(blk->irq >= 0) {
                        res++;
                        res->start = res->end = blk->irq + data->irq_base;
                        res->flags = IORESOURCE_IRQ | IORESOURCE_PLATFORM_VIRTUAL;
                }
                platform_device_register (sdev);
        }
	return 1;
}

static int t7l66xb_probe(struct device *dev)
{
	struct platform_device *pdev = to_platform_device (dev);
	unsigned long pbase = (unsigned long)pdev->resource[0].start;
	unsigned long plen = pdev->resource[0].end - pdev->resource[0].start;
	int err = -ENOMEM;
	struct t7l66xb_data *data;

	data = kmalloc (sizeof (struct t7l66xb_data), GFP_KERNEL);
	if (!data)
		goto out;

	data->irq_base = alloc_irq_space (T7L66XB_NR_IRQS);
	data->irq_nr   = pdev->resource[1].start;

	if (data->irq_base == -1)
		goto out_free_data;

	if(!t7l66xb_add_devices(data, dev, pbase))
		goto out_free_irqs;

	data->mapbase = ioremap(pbase, plen);

	if(!data->mapbase)
		goto out_free_devices;

	dev_set_drvdata(dev, data);
	t7l66xb_setup_irq(data);
	t7l66xb_hwinit(dev);
	
	/* Enable our IRQs */
//	writew(0,    data->mapbase + T7L66XB_SYS_IRR);
	writew(0xbf, data->mapbase + T7L66XB_SYS_IMR);
	
	/* Init finished. */
	printk(KERN_INFO "Toshiba T7L66XB: Revision %d found\n",
	       readw(data->mapbase + T7L66XB_SYS_RIDR));
	printk(KERN_INFO "%s @ 0x%08lx using irq %d-%d on irq %d\n",
	       pdev->name, (unsigned long)data->mapbase, data->irq_base,
	       data->irq_base + T7L66XB_NR_IRQS - 1, data->irq_nr);
	return 0;

out_free_devices:
	t7l66xb_remove_devices(data);
out_free_irqs:
	free_irq_space(data->irq_base, T7L66XB_NR_IRQS);
out_free_data:
	kfree(data);
out:
	return err;
}

static int t7l66xb_remove(struct device *dev)
{
	struct t7l66xb_data *tchip = dev_get_drvdata(dev);
	int i;

	/* Free the subdevice resources */
	for (i = 0; i < tchip->ndevices; i++) {
		platform_device_unregister (&tchip->devices[i]);
		kfree (tchip->devices[i].resource);
	}

	/* Take down IRQ handling */
        for (i = 0; i < T7L66XB_NR_IRQS; i++) {
                int irq = i + tchip->irq_base;
                irq_desc[irq].valid    = 0;
                irq_desc[irq].probe_ok = 0;
                set_irq_handler (irq, NULL);
                set_irq_chip (irq, NULL);
                set_irq_chipdata (irq, NULL);
        }
        set_irq_chained_handler (tchip->irq_nr, NULL);
	// Free IRQ ?

	//FIXME - put chip to sleep?
	
	/* Free core resources */
	iounmap (tchip->mapbase);
	t7l66xb_remove_devices(tchip);
	free_irq_space (tchip->irq_base, T7L66XB_NR_IRQS);
	kfree (tchip);

	return 0;
}


static struct device_driver t7l66xb_device_driver = {
	.name		= "t7l66xb",
	.bus		= &platform_bus_type,
	.probe		= t7l66xb_probe,
	.remove		= t7l66xb_remove,
#ifdef CONFIG_PM
	.suspend	= t7l66xb_suspend,
	.resume		= t7l66xb_resume,
#endif
};


static int __init t7l66xb_init(void)
{
	int retval = 0;
	retval = driver_register (&t7l66xb_device_driver);
	return retval;
}

static void __exit t7l66xb_exit(void)
{
	driver_unregister(&t7l66xb_device_driver);
}

module_init(t7l66xb_init);
module_exit(t7l66xb_exit);

MODULE_DESCRIPTION("Toshiba T7L66XB core driver");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Dirk Opfer");
