/*
 * PCMCIA Support for B4 (pxa27x)
 *
 * 2007 STC Metrotek, Saint-Petersburg, Russia 
 *
 * Author: Pavel Kurochkin <pk@metrotek.spb.ru>
 *
 * Based on pxa2xx_b4.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/device.h>
#include <linux/platform_device.h>

#include <pcmcia/ss.h>

#include <asm/hardware.h>
#include <asm/irq.h>

#include <asm/arch/pxa-regs.h>
#include <asm/arch/b4-gpio.h>

#include "soc_common.h"

static struct pcmcia_irqs irqs[] = {
  { 0, B4_IRQ_GPIO_CF_DETECT, "PCMCIA0 CD" },
//  { 0, B4_IRQ_GPIO_CF_STSCHG, "PCMCIA0 STSCHG" }
};

static int b4_pcmcia_hw_init(struct soc_pcmcia_socket *skt)
{
  /*
   * Setup default state of GPIO outputs
   * before we enable them as outputs.
   */
  GPSR(GPIO48_nPOE) =
    GPIO_bit(GPIO48_nPOE) |
    GPIO_bit(GPIO49_nPWE) |
    GPIO_bit(GPIO50_nPIOR) |
    GPIO_bit(GPIO54_nPCE_2) |
    GPIO_bit(GPIO51_nPIOW);
       
  GPSR(GPIO85_nPCE_1) = GPIO_bit(GPIO85_nPCE_1);

  // inactive reset
  GPSR(B4_GPIO_CF_RESET) = GPIO_bit(B4_GPIO_CF_RESET);
  // compact flash power off
  GPCR(B4_GPIO_CF_PWR) = GPIO_bit(B4_GPIO_CF_PWR);

  pxa_gpio_mode(GPIO48_nPOE_MD);
  pxa_gpio_mode(GPIO49_nPWE_MD);
  pxa_gpio_mode(GPIO50_nPIOR_MD);
  pxa_gpio_mode(GPIO51_nPIOW_MD);
  pxa_gpio_mode(GPIO85_nPCE_1_MD);
  pxa_gpio_mode(GPIO54_nPCE_2_MD);
  pxa_gpio_mode(GPIO55_nPREG_MD);
  pxa_gpio_mode(GPIO56_nPWAIT_MD);
  pxa_gpio_mode(GPIO57_nIOIS16_MD);
  pxa_gpio_mode(B4_GPIO_CF_RESET | GPIO_OUT);
  pxa_gpio_mode(B4_GPIO_CF_PWR | GPIO_OUT);
  pxa_gpio_mode(B4_GPIO_CF_PSKTSEL | GPIO_ALT_FN_1_OUT );

  // b4 platform can only connect pcmcia devices
  // to 0th socket
  BUG_ON(skt->nr != 0);

  skt->irq = B4_IRQ_GPIO_CF_IRQ;

  return soc_pcmcia_request_irqs(skt, irqs, ARRAY_SIZE(irqs));
}

static void b4_pcmcia_hw_shutdown(struct soc_pcmcia_socket *skt)
{
  soc_pcmcia_free_irqs(skt, irqs, ARRAY_SIZE(irqs));
}

#define GPLR_BIT(n) (GPLR((n)) & GPIO_bit((n)))
#define GPSR_BIT(n) (GPSR((n)) = GPIO_bit((n)))
#define GPCR_BIT(n) (GPCR((n)) = GPIO_bit((n)))

static void b4_pcmcia_socket_state(struct soc_pcmcia_socket *skt,
            struct pcmcia_state *state)
{
  if(skt->nr == 0){
    state->detect = GPLR_BIT(B4_GPIO_CF_DETECT) ? 0 : 1;
    state->ready  = GPLR_BIT(B4_GPIO_CF_IRQ) ? 1 : 0;
  }
  else{
    // must never get here
  }

  state->bvd1   = 1; // GPLR_BIT(B4_GPIO_CF_STSCHG) ? 1 : 0; /* cf status changed */
  state->bvd2   = 1;  /* not available */
  state->vs_3v  = 1;  /* not available */
  state->vs_Xv  = 0;  /* not available */
  state->wrprot = 0;  /* not available */
}

static int b4_pcmcia_configure_socket(struct soc_pcmcia_socket *skt,
               const socket_state_t *state)
{
  int ret = 0;

  if(state->flags & SS_RESET) {
    if(skt->nr == 0) {
      // reset uses inverted logic (0 - active)
      GPCR_BIT(B4_GPIO_CF_RESET);
    }
  } else {
    if(skt->nr == 0) {
      GPSR_BIT(B4_GPIO_CF_RESET);
    }
  }

  /* Apply socket voltage */
  switch (state->Vcc) {
    case 0:
      if(skt->nr == 0) {
        GPCR_BIT(B4_GPIO_CF_PWR);
      }
      break;
    case 50:
    case 33:
      /* Apply power to socket */
      if(skt->nr == 0) {
        GPSR_BIT(B4_GPIO_CF_PWR);
      }
      break;
    default:
      printk(KERN_ERR "Unsupported Vcc:%d\n", state->Vcc);
      ret = -1;
      break;
  }
  
  return ret;
}

static void b4_pcmcia_socket_init(struct soc_pcmcia_socket *skt)
{
}

static void b4_pcmcia_socket_suspend(struct soc_pcmcia_socket *skt)
{
}

static struct pcmcia_low_level b4_pcmcia_ops = {
  .owner      = THIS_MODULE,
  .hw_init    = b4_pcmcia_hw_init,
  .hw_shutdown    = b4_pcmcia_hw_shutdown,
  .socket_state    = b4_pcmcia_socket_state,
  .configure_socket  = b4_pcmcia_configure_socket,
  .socket_init    = b4_pcmcia_socket_init,
  .socket_suspend    = b4_pcmcia_socket_suspend,
  .nr      = 1,
};

static struct platform_device *b4_pcmcia_device;

static void b4_pcmcia_release(struct device *dev)
{
  kfree(b4_pcmcia_device);
}

static int __init b4_pcmcia_init(void)
{
  int ret;

  b4_pcmcia_device = kmalloc(sizeof(struct platform_device), GFP_KERNEL);
  if (!b4_pcmcia_device)
    return -ENOMEM;

  memset(b4_pcmcia_device, 0, sizeof(*b4_pcmcia_device));
  b4_pcmcia_device->name = "pxa2xx-pcmcia";
  b4_pcmcia_device->dev.platform_data = &b4_pcmcia_ops;
  b4_pcmcia_device->dev.release = b4_pcmcia_release;

  ret = platform_device_register(b4_pcmcia_device);
  if (ret)
    kfree(b4_pcmcia_device);

  return ret;
}

static void __exit b4_pcmcia_exit(void)
{
  platform_device_unregister(b4_pcmcia_device);
}

module_init(b4_pcmcia_init);
module_exit(b4_pcmcia_exit);

MODULE_AUTHOR("Pavel Kurochkin <pk@metrotek.spb.ru>");
MODULE_DESCRIPTION("B4 platform CompactFlash support");
MODULE_LICENSE("GPL");

